/*
 * @(#) JotmRecovery.java 
 *
 * JOTM: Java Open Transaction Manager 
 *
 *
 * This module was originally developed by 
 *
 *  - Bull S.A. as part of the JOnAS application server code released in 
 *    July 1999 (www.bull.com)
 * 
 * --------------------------------------------------------------------------
 *  The original code and portions created by Bull SA are 
 *  Copyright (c) 1999 BULL SA  
 *  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * -Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 
 *
 * -Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * --------------------------------------------------------------------------
 * $Id: JotmRecovery.java,v 1.1 2004/10/29 05:47:14 tonyortiz Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jotm;

import java.nio.ByteBuffer;
import java.util.Vector;
import javax.transaction.xa.XAResource;
// import javax.transaction.xa.Xid;

import org.objectweb.howl.log.xa.XALogRecord;
import org.objectweb.howl.log.xa.XACommittingTx;
import javax.transaction.xa.XAException;

/**
 * 
 * @author Tony Ortiz
 */

/**
 * Thread class used to commit an Xid of an XAResource,
 * XAResource.commit(Xid).
 */

class commitXAResourceXid extends Thread {
    XAResource commitxares = null;
    javax.transaction.xa.Xid commitxid = null;
    
    commitXAResourceXid (XAResource pxares, javax.transaction.xa.Xid pxid) {
        commitxares = pxares;
        commitxid = pxid;
    }
    
    public void run() {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("thread for commitXAResourceXid");
        }
        
        try {
            commitxares.commit(commitxid, false);
        } catch (XAException e) {
            TraceTm.recovery.error("Unable to commit Xid during Recovery " + e.getMessage());
        }
    }
}

/**
 * Thread class used to abort an Xid of an XAResource,
 * XAResource.rollback(Xid).
 */

class abortXAResourceXid extends Thread {
    XAResource abortxares = null;
    javax.transaction.xa.Xid abortxid = null;
    
    abortXAResourceXid (XAResource pxares, javax.transaction.xa.Xid pxid) {
        abortxares = pxares;
        abortxid = pxid;
    }
    
    public void run() {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("thread for abortXAResourceXid");
        }
        
        try {
            abortxares.rollback(abortxid);
        } catch (XAException e) {
            TraceTm.recovery.error("Unable to rollback Xid during Recovery " + e.getMessage());
        }
    }
}
/**
 * Vector used to hold all Resource Managers that were
 * registered when the system crashed, may have XIDs that
 * need to be recovered.
 */

class RecoverRmInfo {

    private String recoverRm = null;
    private byte[] recoverXares = null;
    private int recoverIndex = 0;

    public void addRecoverRmXaRes (String rrmName, byte[] rrmXares, int rrmIndex) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Recover Resource Manager= " + rrmName);
            TraceTm.recovery.debug("Recover XAResource = " + new String(rrmXares));
            TraceTm.recovery.debug("Recover Index = " + rrmIndex);
        }

        recoverRm = rrmName;
        recoverXares = rrmXares;
        recoverIndex = rrmIndex;
    }

    public String getRecoverRm () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverRm= " + recoverRm);
        }
        return recoverRm;
    }

    public byte[] getRecoverXaRes () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverXares= " + new String(recoverXares));
        }
        return recoverXares;
    }
    
    public int getRecoverIndex () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverIncex= " + recoverIndex);
        }
        return recoverIndex;
    }
}

/**
 * Vector and Array used to hold all Transactions that were
 * in a commit pending status when the system crashed, these
 * transactions (XIDs) may need to be recovered.
 */

class TxxidRecovered {

    private int recoverindex = 0;
    private byte[] recoverxares = null;
    private byte[] recoverxid = null;
    private int recoverstatus = 0;
    private int recoveraction = 0;    // 0-Do nothing, implied already committed
                                      // 1-Commit, 2-Rollback

    public void addXidInfo (int pindex, byte[] pxares, byte[] pxid, int pstatus) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Recovered index= " + pindex);
            TraceTm.recovery.debug("Recovered xares= " + new String(pxares));
            TraceTm.recovery.debug("Recovered xid= " + new String(pxid));
            TraceTm.recovery.debug("Recovered status= " + pstatus);
        }

        recoverindex = pindex;
        recoverxares = pxares;
        recoverxid = pxid;
        recoverstatus = pstatus;
    }

    public int getRecoverindex () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverindex= " + recoverindex);
        }
        return recoverindex;
    }

    public byte[] getRecoverxares () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverxares= " + new String(recoverxares));
        }
        return recoverxares;
    }

    public byte[] getRecoverxid () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverXid= " + new String(recoverxid));
        }
        return recoverxid;
    }

    public int getRecoverstatus () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("recoverstatus= " + recoverstatus);
        }
        return recoverstatus;
    }
    
    public int getRecoveraction () {
        return recoveraction;
    }
    
    public void setRecoveraction (int paction) {
        recoveraction = paction;
    }
}


class TxRecovered {
    private long recoverydatetime = 0L;
    private byte[] txxid = null;
    private String txdatetime = null;
    private int xidcount = 0;
    private XACommittingTx xacommittingtx = null;

    private TxxidRecovered [] xidinfo;

    public void addtxrecovered (long prdt, byte[] ptxxid, String ptdt, int pxcnt, XACommittingTx pxacmtx) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Recover tx prdt= " + prdt);
            TraceTm.recovery.debug("Recover tx ptxxid= " + new String(ptxxid));
            TraceTm.recovery.debug("Recover tx ptdt= " + ptdt);
            TraceTm.recovery.debug("Recover tx pxcnt= " + pxcnt);
        }

        recoverydatetime = prdt;
        txxid = ptxxid;
        txdatetime = ptdt;
        xidcount = pxcnt;
        xacommittingtx = pxacmtx;

        xidinfo = new TxxidRecovered [xidcount];
    }

    public void addRecoverTxXidInfo (TxxidRecovered ptxxidr, int rindx) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("addRecoverTxXidInfo");
        }
        xidinfo[rindx] = ptxxidr;
    }
    
    public int getxidcount () {
        return xidcount;
    }
    
    public TxxidRecovered getRecoverTxXidInfo (int rindx) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("getRecoverTxXidInfo");
        }
        return xidinfo[rindx];
    }
    
    public XACommittingTx getXACommittingTx () {
        return xacommittingtx;
    }
}

public class JotmRecovery {
    
    private static JotmRecovery unique = null;
    private static Vector xaRecover = new Vector ();
    private static Vector rmivec = new Vector ();
    private Vector userRecoveryRecords = new Vector ();
    
    /**
     * Constructor.
     */

    public JotmRecovery(){

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("JotmRecovery constructor");
        }

        unique = this;
    }
    
    /**
     * Returns the unique instance of the class or <code>null</code> if not
     * initialized.
     *
     * @return The <code>JotmRecovery</code> object created 
     */

    public static JotmRecovery getJotmRecovery() {

        return unique;
    }
    
    /**
     * Returns the index of the Resource Manager's XAResource.
     *
     * @return Index of the Resource Manager's XAResource. 
     */

    public int getRmIndex(byte [] pxares) {
        int numRm = rmivec.size();
        RecoverRmInfo myrecoverRmInfo;
        
        for (int i=0; i<numRm; i++){
            myrecoverRmInfo = (RecoverRmInfo) rmivec.elementAt(i);
            byte [] inrmxares = myrecoverRmInfo.getRecoverXaRes();
            
            if ( TraceTm.recovery.isDebugEnabled() ) {
                TraceTm.recovery.debug("XAResource param " + pxares);
                TraceTm.recovery.debug("XAResource in rm " + inrmxares);
            }

            if (inrmxares == pxares) {
                return myrecoverRmInfo.getRecoverIndex();
            }
        }    
        return 99;
    }

    public Vector getUserRecoveryVector() { // currently used by test suite
        return userRecoveryRecords;
    }
    
    /**
     * Processes an XACOMMIT entry (putCommit) that does not have an associated
     * XADONE entry (putDone).
     *
     * @param lr LogRecord that was passed to onRecord() method.
     */

    public boolean rebuildTransaction (XALogRecord lr) {

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("rebuildTransaction");
        }
        
        RecoverRmInfo myrecoverRmInfo = null;
        TxxidRecovered myrecoverTxInfo = null;
        TxRecovered mytxRecovered = null;
        
        // Temporary Recovery Record
        byte [] tempRec;
        byte [] rt = new byte [3];  // Applies to all Records

        // Resource Manager Record Type 1 fields
        long rmdatetime;
        int rmcount;

        // Resource Manager Record Type 2
        byte [] resmgr2;

        // Resource Manager Record Type 2 fields
        int rmlength;
        byte [] rmname = null;
        int rmxareslength;
        byte [] rmxares = null;
        int rmindx;

        // Recovery Record Type 1 fields
        long rcdatetime;
        int txxidlength;
        byte [] txXid = null;
        int txdatelength;
        byte [] txdatetime;
        int xarescount = 0;

        // Check if there already exist a recovery record in our array with the
        // same txXid but a older datetime. If so remove it and store this one,
        // otherwise throw this one away.

        // Recovery Record Type 2
        byte [] recov2;

        // Recovery Record Type 2 fields
        byte [] rt2 = new byte [3];
        int xaresindex;
        int xareslength;
        byte [] xares;
        int xidlength;
        byte [] recoveryxid;
        int xidstatus;

        String trt;

        tempRec = lr.getFields() [0];
        XACommittingTx myxacommittx = lr.getTx();

        ByteBuffer rr = ByteBuffer.wrap(tempRec);

        rr.get(rt, 0, 3);
        trt = new String (rt);

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("Recovery Record type= " + trt);
        }

        if (trt.equals("RM1")) {

            rmdatetime = rr.getLong();
            rmcount = rr.getInt();

            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("Resource Manager count= " + rmcount);
            }

            for (int i=1; i <= rmcount; i++) {
                resmgr2 = lr.getFields()[i];

                ByteBuffer resm2 = ByteBuffer.wrap(resmgr2);
                resm2.get(rt2, 0, 3);
                trt = new String (rt2);

                if (trt.equals("RM2")) {
                    rmlength = resm2.getInt();
                    rmname = new byte[rmlength];
                    resm2.get(rmname, 0, rmlength);
                    xareslength = resm2.getInt();
                    xares = new byte[xareslength];
                    resm2.get(xares, 0, xareslength);
                    rmindx = resm2.getInt();

                    String myrmname = new String(rmname);
                    String myxares = new String(xares);
                    myrecoverRmInfo = new RecoverRmInfo();

                    myrecoverRmInfo.addRecoverRmXaRes (myrmname, xares, rmindx);
                    rmivec.addElement(myrecoverRmInfo);
                }
            }
        } 
        else if (trt.equals("RR1")) {
                rcdatetime = rr.getLong();
                txxidlength = rr.getInt();
                txXid = new byte[txxidlength];
                rr.get(txXid, 0, txxidlength);
                txdatelength = rr.getInt();
                txdatetime = new byte[txdatelength];
                rr.get(txdatetime, 0, txdatelength);
                xarescount = rr.getInt();

                String mytxXid = new String(txXid);
                String mytxdatetime = new String(txdatetime);

                mytxRecovered = new TxRecovered();

                if (TraceTm.recovery.isDebugEnabled()) {
                    TraceTm.recovery.debug("XAResource count= " + xarescount);
                }

                mytxRecovered.addtxrecovered (rcdatetime, txXid, mytxdatetime, xarescount, myxacommittx);
                
                for (int i=1; i <= xarescount; i++) {
                    myrecoverTxInfo = new TxxidRecovered();
                    recov2 = lr.getFields()[i];

                    ByteBuffer rr2 = ByteBuffer.wrap(recov2);
                    rr2.get(rt2, 0, 3);
                    trt = new String (rt2);

                    if (trt.equals("RR2")) {
                        xaresindex = rr2.getInt();
                        xareslength = rr2.getInt();
                        xares = new byte[xareslength];
                        rr2.get(xares, 0, xareslength);
                        xidlength = rr2.getInt();
                        recoveryxid = new byte[xidlength];
                        rr2.get(recoveryxid, 0, xidlength);
                        xidstatus = rr2.getInt();

                        String myxares = new String(xares);
                        String myrecoveryxid = new String(recoveryxid);

                        myrecoverTxInfo.addXidInfo (xaresindex, xares, recoveryxid, xidstatus);
                        mytxRecovered.addRecoverTxXidInfo (myrecoverTxInfo, i-1);
                    }
                }
                xaRecover.addElement(mytxRecovered);
        } 
        else if (trt.equals("RU1")) {
                XidImpl.setUuids( rr.getLong(), rr.getLong());
        } 
        else {
            if (TraceTm.recovery.isDebugEnabled()) {
                TraceTm.recovery.debug("Unknown record type during replay = " + trt);
            }
            // add this byteBuffer of this record to a vector for user recovery
            rr.rewind();
            userRecoveryRecords.add( rr );
            userRecoveryRecords.add( myxacommittx );
            return false;
        }
        return true;
    }

    public void recoverTransactions (Vector rmreg) throws XAException {
        if ( TraceTm.recovery.isDebugEnabled() ) {
            TraceTm.recovery.debug("recoverTransactions");
        }

        RmRegistration myregisteredRmInfo = null;
        String myregrm = null;
        XAResource myregxares = null;
        int rmregsize = rmreg.size();  // number of Resource Managers registered
        
        RecoverRmInfo myrecoverRmInfo = null;
        String myrm = null;
        byte[] byxares = null;

        int rcflag = 0;

        // read the tmRecovered object and determine if any of the xaResources
        // require recovery (xares.recover call)

        int rmsize = rmivec.size();    // number of Resource Managers when system crashed

        if (rmsize == 0) {
            if ( TraceTm.recovery.isDebugEnabled() ) {
                TraceTm.recovery.debug("Nothing to recover");
            }
            return;
        }

        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("number Resource Manager recover= " + rmsize);
        }

        // we can only recover Resource Managers that are in our log
        // (RecoverRmInfo) and have been registered (RmRegistered)
        
        for (int i=0; i<rmregsize; i++) {
            
            RmRegistration myrmreg = (RmRegistration) rmreg.elementAt(i);
            myregrm = myrmreg.rmGetName();
            myregxares = myrmreg.rmGetXaRes();
            
            if ( TraceTm.recovery.isDebugEnabled() ) {
                TraceTm.recovery.debug("Registered Resource Manager " + myregrm);
                TraceTm.recovery.debug("Registered XAResource " + myregxares);
            }

            for (int j=0; j<rmsize; j++) {
                myrecoverRmInfo = (RecoverRmInfo) rmivec.elementAt(j);
                myrm = myrecoverRmInfo.getRecoverRm();
                byxares = myrecoverRmInfo.getRecoverXaRes();
                
                if ( TraceTm.recovery.isDebugEnabled() ) {
                    TraceTm.recovery.debug("Recover Resource Manager " + myrm);
                    TraceTm.recovery.debug("Recover XAResource " + new String(byxares));
                }
                
                if (myregrm.equals(myrm)){
                    javax.transaction.xa.Xid javaxid [] = new javax.transaction.xa.Xid[100];
                    
                    try {
                        // recover returns javax.transaction.xa.Xid objects
                        // we must cast to org.objectweb.jotm.Xid objects
                        javaxid = myregxares.recover(rcflag);
                        
                        if ( TraceTm.recovery.isDebugEnabled() ) {
                            TraceTm.recovery.debug("javaxid size= " + javaxid.length);
                        }
                    } catch (XAException e) {
                        throw new XAException("xaResource.recover call failed during recovery " + e.getMessage());
                    } catch (Exception e) {
                        if ( TraceTm.recovery.isDebugEnabled() ) {
                            TraceTm.recovery.debug("Cannot recover resource " + e.getMessage());
                        }
                        break;
                    }
                    Xid rcxid [] = new Xid[javaxid.length];
                    
                    for (int xx=0; xx<javaxid.length; xx++) {
                       rcxid[xx]= (Xid)javaxid[xx];
                    }

                    if (rcxid.length == 0) {
                        if(TraceTm.recovery.isDebugEnabled()) {
                            TraceTm.recovery.debug("No XIDs to recover for Xares= "+ myregxares);
                        }
                        break;
                    }
                    
                    // set the action required for the XIDs located in our
                    // txxidRecovered (array), pointed to by TxRecovered (vector)
                    settxxidrecoveraction (myregxares, rcxid);
                    break;
                }
            }
        }
        
        // We can now walk through TxRecovered and perform the action on
        // each of the Xids stored in TxxidRecovered.
        
        // doActionXidRecover ();
        
        // We can now walk through TxRecovered and delete (howl done) any TxRecovered
        // entries that have had all their Xids (TxxidRecovered) committed or rolled
        // back successfully.
        // If any Xids have not been committed or rolledback, create a new TxRecovered
        // entry (howl commit) with the Xids (TxxidRecovered) that are still pending.
        
        doCleanupXidRecover ();
    }

    private void settxxidrecoveraction (XAResource actxares, Xid [] actionxid) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("settxxidrecoveraction");
        }
        
        TxxidRecovered myrecoverTxInfo = null;
        TxRecovered mytxRecovered = null;

        for (int i=0; i<actionxid.length; i++) {
            boolean xidfound = false;
            Xid myxid = actionxid[i];
            byte [] mybrqu = myxid.getBranchQualifier();
            
            if (myxid.IsThisOneOfOurs(mybrqu)){
                
                for (int j=0; j<xaRecover.size(); j++){
                    mytxRecovered = (TxRecovered) xaRecover.elementAt(j);
                    
                    for (int k=0; k<mytxRecovered.getxidcount(); k++){
                        myrecoverTxInfo = mytxRecovered.getRecoverTxXidInfo(k);
                        byte [] mytxxid = myrecoverTxInfo.getRecoverxid();
                        
                        if (TraceTm.recovery.isDebugEnabled()) {
                            TraceTm.recovery.debug("mytxxid= " + new String(mytxxid));
                            TraceTm.recovery.debug("myxid  = " + myxid.toString(true));
                        }
                        
                        if ((new String(mytxxid)).equals(myxid.toString(true))){
                            xidfound = true;
                            int myaction = 1;    // Commit
                            myrecoverTxInfo.setRecoveraction(myaction);
                            mytxRecovered.addRecoverTxXidInfo(myrecoverTxInfo,k);
                            commitresource(actxares, myxid);
                            break;
                        }
                    }
                    if (xidfound) {    // xid found in a TxxidRecovered of the TxRecovered vector
                        break;
                    }
                }
                if (!xidfound) {    // xid not found in any TxxidRecovered of the TxRecovered vector
                    abortimmediate (actxares, myxid);
                }
            } else {
                if(TraceTm.recovery.isDebugEnabled()) {
                    TraceTm.recovery.debug("Xid is not one of ours");
                }
                // return;
            }
        }
    }

    private void doActionXidRecover () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("doActionXidRecover");
        }
        
        TxxidRecovered myrecoverTxInfo = null;
        TxRecovered mytxRecovered = null;
        
        for (int i=0; i<xaRecover.size(); i++){
            mytxRecovered = (TxRecovered) xaRecover.elementAt(i);
            // myrecoverTxInfo = mytxRecovered.getRecoverTxXidInfo(i);
            
            for (int j=0; j<mytxRecovered.getxidcount(); j++){
                myrecoverTxInfo = mytxRecovered.getRecoverTxXidInfo(j);

                // If action = 0; XAResource is known but the XID associated
                // to is was not returned in the xares.recover call. This implies
                // that the XID was already committed, just ignore.
                // If action = 1; XAResource is known and the xares.recover call
                // returned the specifed XID. This implies that we should attempt
                // to commit the XID (if possible).
                
                if (myrecoverTxInfo.getRecoveraction() == 1) {
                    byte [] myrcxid = myrecoverTxInfo.getRecoverxid();
                    byte [] myxares = myrecoverTxInfo.getRecoverxares();
                    XAResource usexares = (XAResource) ((Object) myxares);
                    Xid myxid = (Xid) ((Object) myrcxid);
                    javax.transaction.xa.Xid usexid = (javax.transaction.xa.Xid) myxid;

                    commitresource(usexares, myxid);
                    
                }
            }
        }
    }

    private void commitresource (XAResource commitxares, Xid commitxid) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("commitresource");
        }

        new commitXAResourceXid (commitxares, commitxid).start();
    }

    private void abortimmediate (XAResource abortxares, Xid abortxid) {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("abortimmediate");
        }

        new abortXAResourceXid (abortxares, abortxid).start();
    }
    
    private void doCleanupXidRecover () {
        if (TraceTm.recovery.isDebugEnabled()) {
            TraceTm.recovery.debug("doCleanupXidRecover");
        }
        
        TxxidRecovered myrecoverTxInfo = null;
        TxRecovered mytxRecovered = null;

        
        for (int i=0; i<xaRecover.size(); i++){
            boolean possibleheuristic = false;
            XACommittingTx myxacommittingtx = null;
            mytxRecovered = (TxRecovered) xaRecover.elementAt(i);
            myrecoverTxInfo = mytxRecovered.getRecoverTxXidInfo(i);
            
            for (int j=0; j<mytxRecovered.getxidcount(); j++){

                // If action = 0; XAResource is known but the XID associated
                // to is was not returned in the xares.recover call. This implies
                // that the XID was already committed, just ignore.
                // If action = 1; XAResource is known and the xares.recover call
                // returned the specifed XID. This implies that we should attempt
                // to commit the XID (if possible).
                
                if (myrecoverTxInfo.getRecoveraction() == 9) {
                    possibleheuristic = true;
                }
            }
            if (possibleheuristic) {
                if (TraceTm.recovery.isDebugEnabled()) {
                    TraceTm.recovery.debug("build new txRecovered entry");
                }
            } else {
                if (TraceTm.recovery.isDebugEnabled()) {
                    TraceTm.recovery.debug("remove txRecovered entry");
                }
                
            	myxacommittingtx = mytxRecovered.getXACommittingTx();
                byte [] rmDone = new byte [11];
                byte [] [] rmDoneRecord = new byte [1] [11];

                rmDone = "RR3JOTMDONE".getBytes();

                try {
                    rmDoneRecord [0] = rmDone;
                    TransactionRecoveryImpl.getTransactionRecovery().howlDoneLog (rmDoneRecord, myxacommittingtx);
                } catch (Exception f) {
                    String howlerror =
                        "Cannot howlDoneLog:"
                        + f
                        + "--"
                        + f.getMessage();
                    TraceTm.jotm.error("Got LogException from howlDoneLog: "+ howlerror);
                }
            }
        }
    }
}